"""This module defines some classes to make PyTorch metric compatible
with ``VanillaModel``."""

from __future__ import annotations

__all__ = ["PaddedSequenceMetric", "VanillaMetric"]


import torch
from coola.utils import str_mapping
from torch import Tensor

from gravitorch import constants as ct
from gravitorch.engines.base import BaseEngine
from gravitorch.models.metrics.base import BaseMetric


class VanillaMetric(BaseMetric):
    r"""Implements a wrapper to make compatible most of the metrics with
    ``gravitorch.models.VanillaModel``.

    This class works for every ``BaseMetric`` instance that has two
    inputs: prediction and target.

    Args:
    ----
        metric (``BaseMetric`` or dict): Specifies the metric function
            or its configuration.
        mode (str, optional): Specifies the mode (e.g train or eval).
        prediction_key (str, optional): Specifies the key where the
            prediction is if the network output is a dict.
            Default: ``prediction``
        target_key (str, optional): Specifies the key where the target
            is in the batch input. Default: ``target``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.models.metrics import TopKAccuracy
        >>> # Initialization with a metric object.
        >>> metric = VanillaMetric(TopKAccuracy(mode="train", topk=[1]))
        >>> metric
        VanillaMetric(
          (prediction_key): prediction
          (target_key): target
          (metric): TopKAccuracy(
            (mode): train
            (name): acc_top
            (topk): (1,)
            (states):
              (1): AccuracyState(num_predictions=0)
          )
        )
        >>> # Initialization with the config of a metric.
        >>> metric = VanillaMetric(mode="eval", metric={"_target_": "TopKAccuracy", "topk": [1]})
        >>> metric
        VanillaMetric(
          (prediction_key): prediction
          (target_key): target
          (metric): TopKAccuracy(
            (mode): eval
            (name): acc_top
            (topk): (1,)
            (states):
              (1): AccuracyState(num_predictions=0)
          )
        )
        >>> # Customize keys.
        >>> net_out = {"next_sentence_prediction": ...}
        >>> batch = {"next_sentence_target": ...}
        >>> metric = VanillaMetric(
        ...     TopKAccuracy(mode="train", topk=[1]),
        ...     prediction_key="next_sentence_prediction",
        ...     target_key="next_sentence_target",
        ... )
        >>> metric
        VanillaMetric(
          (prediction_key): next_sentence_prediction
          (target_key): next_sentence_target
          (metric): TopKAccuracy(
            (mode): train
            (name): acc_top
            (topk): (1,)
            (states):
              (1): AccuracyState(num_predictions=0)
          )
        )
    """

    def __init__(
        self,
        metric: BaseMetric | dict,
        mode: str | None = None,
        prediction_key: str = ct.PREDICTION,
        target_key: str = ct.TARGET,
    ) -> None:
        super().__init__()
        self.metric = self._setup_metric(metric=metric, mode=mode)
        self._prediction_key = prediction_key
        self._target_key = target_key

    def attach(self, engine: BaseEngine) -> None:
        self.metric.attach(engine)

    def extra_repr(self) -> str:
        return str_mapping({"prediction_key": self._prediction_key, "target_key": self._target_key})

    def forward(self, cri_out: dict, net_out: dict, batch: dict) -> dict | None:
        r"""Updates the metric given a mini-batch of examples.

        Args:
        ----
            cri_out (dict): Specifies the criterion output. (not used
                in this implementation)
            net_out (dict or ``torch.Tensor``): Specifies the network
                output which contains the prediction.
            batch (dict): Specifies the batch which contains the
                target.

        Returns:
        -------
            dict or ``None``: A dict with the metric values. ``None``
                means no metric value is returned.
        """
        return self.metric(
            self._get_prediction_from_net_out(net_out),
            self._get_target_from_batch(batch),
        )

    def reset(self) -> None:
        self.metric.reset()

    def value(self, engine: BaseEngine | None = None) -> dict:
        return self.metric.value(engine)

    def _get_prediction_from_net_out(self, net_out: dict) -> Tensor:
        r"""Gets the prediction from the network output.

        Args:
        ----
            net_out (dict): Specifies the network output which
                contains the prediction.

        Returns:
        -------
            ``torch.Tensor``: The tensor of predictions.
        """
        return net_out[self._prediction_key].detach()

    def _get_target_from_batch(self, batch: dict) -> Tensor:
        r"""Gets the target from the batch. The target is the tensor with
        the key 'target'.

        Args:
        ----
            batch (dict): Specifies the batch which contains the target.

        Returns:
        -------
            ``torch.Tensor``: The tensor of targets.
        """
        return batch[self._target_key].detach()

    def _setup_metric(self, metric: BaseMetric | dict, mode: str | None) -> BaseMetric:
        r"""Sets up the metric.

        If the input is a dict containing the configuration of the
        metric, it will instantiate the metric given its config by
        using the ``factory`` method of ``BaseMetric``.

        Args:
        ----
            metric (``BaseMetric`` or dict): Specifies the metric
                function or its configuration.
            mode (str, optional): Specifies the mode (e.g train or
                eval).

        Returns:
        -------
            ``BaseMetric``: an instance of the metric.
        """
        if isinstance(metric, dict):
            if mode is not None:
                metric["mode"] = mode
            metric = BaseMetric.factory(**metric)
        return metric


class PaddedSequenceMetric(VanillaMetric):
    r"""Implements a wrapper to make compatible most of the metrics with
    ``gravitorch.models.VanillaModel``.

    This metric is designed to work on sequences. This metric should
    have at least two inputs:

        - the prediction which is a ``torch.Tensor`` of shape
            ``(sequence_length, batch_size, *)`` or
            ``(batch_size, sequence_length, *)`` where ``*`` means
            any number of additional dimensions. This tensor is
            converted to a tensor of shape
            ``(sequence_length * batch_size, *)`` and then feeds to
            the wrapped metric.
        - the target which is a ``torch.Tensor`` of shape
            ``(sequence_length, batch_size, *)`` or
            ``(batch_size, sequence_length, *)`` where ``*`` means
            any number of additional dimensions.
            This tensor is converted to a tensor of shape
            ``(sequence_length * batch_size, *)`` and then feeds to
            the wrapped metric.

    The input mask is optional. If no mask is provided, all the steps
    are considered as valid. The mask is a ``torch.Tensor`` of shape
    ``(sequence_length, batch_size)`` or
    ``(batch_size, sequence_length)``. The type of the tensor can be
    ``torch.int`` or``torch.long`` or``torch.float`` or ``torch.bool``
    with the following values:

        - valid value: ``True`` or ``1`` if ``valid_value=True``,
            otherwise ``False`` or ``0``.
        - invalid value: ``False`` or ``0`` if ``valid_value=True``,
            otherwise ``True`` or ``1``.

    Args:
    ----
        metric (``BaseMetric`` or dict): Specifies the metric function
            or its configuration.
        mode (str, optional): Specifies the mode (e.g train or eval).
        prediction_key (str, optional): Specifies the key where the
            prediction is if the network output is a dict.
            Default: ``prediction``
        target_key (str, optional): Specifies the key where the target
            is in the batch input. Default: ``target``
        mask_key (str, optional): Specifies the key where the target
            is in the batch input. Default: ``target``
        valid_value (bool, optional): Indicates the valid values in
            the mask. If ``True``, the valid values are indicated by
            a ``True`` in the mask. If ``False``, the valid values are
            indicated by a ``False`` in the mask. Default: ``True``
        mask_in_batch (bool, optional): Indicates if the mask is in
            ``batch`` or ``net_out``. If ``True``, the mask is taken
            from the input ``batch``, otherwise it is taken from the
            input ``net_out``. Default: ``True``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.models.metrics import PaddedSequenceMetric, AbsoluteError
        >>> metric = PaddedSequenceMetric(AbsoluteError(mode="train"))
        >>> metric(
        ...     cri_out={},
        ...     net_out={"prediction": torch.ones(2, 6, 1)},
        ...     batch={"target": torch.zeros(2, 6, 1)},
        ... )
        >>> metric.value()
        {'train/abs_err_mean': 1.0, 'train/abs_err_min': 1.0, 'train/abs_err_max': 1.0, 'train/abs_err_sum': 12.0, 'train/abs_err_num_predictions': 12}
        >>> metric(
        ...     cri_out={},
        ...     net_out={"prediction": torch.ones(2, 4, 1)},
        ...     batch={
        ...         "target": torch.zeros(2, 4, 1),
        ...         "mask": torch.tensor([[True, True, False, True], [True, True, True, False]]),
        ...     },
        ... )
        >>> metric.value()
        {'train/abs_err_mean': 1.0, 'train/abs_err_min': 1.0, 'train/abs_err_max': 1.0, 'train/abs_err_sum': 18.0, 'train/abs_err_num_predictions': 18}
    """

    def __init__(
        self,
        metric: BaseMetric | dict,
        mode: str | None = None,
        prediction_key: str = ct.PREDICTION,
        target_key: str = ct.TARGET,
        mask_key: str = ct.MASK,
        valid_value: bool = True,
        mask_in_batch: bool = True,
    ) -> None:
        super().__init__(
            metric=metric,
            mode=mode,
            prediction_key=prediction_key,
            target_key=target_key,
        )
        self._mask_key = mask_key
        self._valid_value = bool(valid_value)
        self._mask_in_batch = bool(mask_in_batch)

    def extra_repr(self) -> str:
        return str_mapping(
            {
                "prediction_key": self._prediction_key,
                "target_key": self._target_key,
                "mask_key": self._mask_key,
                "valid_value": self._valid_value,
                "mask_in_batch": self._mask_in_batch,
            }
        )

    def forward(self, cri_out: dict, net_out: dict, batch: dict) -> dict | None:
        r"""Updates the metric given a mini-batch of examples.

        Args:
        ----
            cri_out (dict): Specifies the criterion output. (not used
                in this implementation)
            net_out (dict): Specifies the network output which
                contains the prediction.
            batch (dict): Specifies the batch which contains the
                target and the mask.

        Returns:
        -------
            dict or ``None``: A dict with the metric values. ``None``
                means no metric value is returned.
        """
        prediction = self._get_prediction_from_net_out(net_out)
        target = self._get_target_from_batch(batch)

        # See the batch of sequences as a batch of examples
        prediction = prediction.view(-1, *prediction.shape[2:])
        target = target.view(-1, *target.shape[2:])

        # Get the mask and remove the examples that are masked
        mask = (
            batch.get(self._mask_key, None)
            if self._mask_in_batch
            else net_out.get(self._mask_key, None)
        )
        if mask is not None:
            mask = mask.view(-1).bool()
            if not self._valid_value:
                mask = torch.logical_not(mask)
            prediction = prediction[mask]
            target = target[mask]

        return self.metric(prediction, target)



from __future__ import annotations

__all__ = ["BaseCoreCreator", "is_core_creator_config", "setup_core_creator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config
from torch.nn import Module
from torch.optim import Optimizer

from gravitorch.datasources.base import BaseDataSource
from gravitorch.lr_schedulers.base import LRSchedulerType
from gravitorch.utils.format import str_target_object

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class BaseCoreCreator(ABC, metaclass=AbstractFactory):
    """Defines the base class to create some core engine modules.

    In MLTorch, the core engine modules are:

        - datasource
        - model
        - optimizer
        - LR scheduler

    Note it is possible to create these core modules without using
    this class.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine
        >>> from gravitorch.creators.core import CoreCreator
        >>> creator = CoreCreator(
        ...     datasource={"_target_": "gravitorch.testing.DummyDataSource"},
        ...     model={"_target_": "gravitorch.testing.DummyClassificationModel"},
        ...     optimizer={"_target_": "torch.optim.SGD", "lr": 0.01},
        ...     lr_scheduler={"_target_": "torch.optim.lr_scheduler.StepLR", "step_size": 5},
        ... )
        >>> creator  # doctest: +ELLIPSIS
        CoreCreator(
          (datasource): DummyDataSource(
              (datasets):
                (train): DummyDataset(num_examples=4, feature_size=4)
                (eval): DummyDataset(num_examples=4, feature_size=4)
              (dataloader_creators):
                (train): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
                (eval): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
            )
          (model): DummyClassificationModel(
              (linear): Linear(in_features=4, out_features=3, bias=True)
              (criterion): CrossEntropyLoss()
            )
          (optimizer): SGD (
            Parameter Group 0...
                lr: 0.01
                maximize: False
                momentum: 0
                nesterov: False
                weight_decay: 0
            )
          (lr_scheduler): <torch.optim.lr_scheduler.StepLR object at 0x...>
        )
        >>> engine = create_dummy_engine()
        >>> datasource, model, optimizer, lr_scheduler = creator.create(engine)
        >>> datasource
        DummyDataSource(
          (datasets):
            (train): DummyDataset(num_examples=4, feature_size=4)
            (eval): DummyDataset(num_examples=4, feature_size=4)
          (dataloader_creators):
            (train): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
            (eval): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
        )
        >>> model
        DummyClassificationModel(
          (linear): Linear(in_features=4, out_features=3, bias=True)
          (criterion): CrossEntropyLoss()
        )
        >>> optimizer  # doctest: +ELLIPSIS
        SGD (
          Parameter Group 0...
              lr: 0.01
              maximize: False
              momentum: 0
              nesterov: False
              weight_decay: 0
        )
        >>> lr_scheduler  # doctest: +ELLIPSIS
        <torch.optim.lr_scheduler.StepLR object at 0x...>
    """

    @abstractmethod
    def create(
        self, engine: BaseEngine
    ) -> tuple[BaseDataSource, Module, Optimizer | None, LRSchedulerType | None]:
        r"""Creates the core engine modules.

        This method is responsible to register the event handlers
        associated to the core engine modules.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine.

        Returns:
        -------
            tuple with 4 values with the following structure:
                - ``gravitorch.datasources.BaseDataSource``: The
                    initialized datasource.
                - ``torch.nn.Module``: The instantiated model.
                - ``torch.optim.Optimizer`` or ``None``: The
                    instantiated optimizer or ``None`` if there
                    is no optimizer (evaluation mode only).
                - ``LRSchedulerType`` or ``None``: The instantiated
                    learning rate (LR) scheduler or ``None`` if
                    there is no learning rate scheduler.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.testing import create_dummy_engine
            >>> from gravitorch.creators.core import CoreCreator
            >>> creator = CoreCreator(
            ...     datasource={"_target_": "gravitorch.testing.DummyDataSource"},
            ...     model={"_target_": "gravitorch.testing.DummyClassificationModel"},
            ...     optimizer={"_target_": "torch.optim.SGD", "lr": 0.01},
            ...     lr_scheduler={"_target_": "torch.optim.lr_scheduler.StepLR", "step_size": 5},
            ... )
            >>> engine = create_dummy_engine()
            >>> datasource, model, optimizer, lr_scheduler = creator.create(engine)
            >>> datasource
            DummyDataSource(
              (datasets):
                (train): DummyDataset(num_examples=4, feature_size=4)
                (eval): DummyDataset(num_examples=4, feature_size=4)
              (dataloader_creators):
                (train): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
                (eval): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
            )
            >>> model
            DummyClassificationModel(
              (linear): Linear(in_features=4, out_features=3, bias=True)
              (criterion): CrossEntropyLoss()
            )
            >>> optimizer  # doctest: +ELLIPSIS
            SGD (
              Parameter Group 0...
                  lr: 0.01
                  maximize: False
                  momentum: 0
                  nesterov: False
                  weight_decay: 0
            )
            >>> lr_scheduler  # doctest: +ELLIPSIS
            <torch.optim.lr_scheduler.StepLR object at 0x...>
        """


def is_core_creator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseCoreCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseCoreCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.core import is_core_creator_config
        >>> is_core_creator_config(
        ...     {
        ...         "_target_": "gravitorch.creators.core.CoreCreator",
        ...         "datasource": {"_target_": "gravitorch.testing.DummyDataSource"},
        ...         "model": {"_target_": "gravitorch.testing.DummyClassificationModel"},
        ...         "optimizer": {"_target_": "torch.optim.SGD", "lr": 0.01},
        ...         "lr_scheduler": {"_target_": "torch.optim.lr_scheduler.StepLR", "step_size": 5},
        ...     }
        ... )
        True
    """
    return is_object_config(config, BaseCoreCreator)


def setup_core_creator(creator: BaseCoreCreator | dict) -> BaseCoreCreator:
    r"""Sets up the core engine modules creator.

    The core engine modules creator is instantiated from its
    configuration by using the ``BaseCoreModulesCreator`` factory
    function.

    Args:
    ----
        creator (``BaseCoreCreator`` or dict): Specifies the
            core engine modules creator or its configuration.

    Returns:
    -------
        ``BaseCoreCreator``: The instantiated core engine
            modules creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.core import setup_core_creator
        >>> creator = setup_core_creator(
        ...     {
        ...         "_target_": "gravitorch.creators.core.CoreCreator",
        ...         "datasource": {"_target_": "gravitorch.testing.DummyDataSource"},
        ...         "model": {"_target_": "gravitorch.testing.DummyClassificationModel"},
        ...         "optimizer": {"_target_": "torch.optim.SGD", "lr": 0.01},
        ...         "lr_scheduler": {"_target_": "torch.optim.lr_scheduler.StepLR", "step_size": 5},
        ...     }
        ... )
        >>> creator  # doctest: +ELLIPSIS
        CoreCreator(
          (datasource): DummyDataSource(
              (datasets):
                (train): DummyDataset(num_examples=4, feature_size=4)
                (eval): DummyDataset(num_examples=4, feature_size=4)
              (dataloader_creators):
                (train): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
                (eval): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
            )
          (model): DummyClassificationModel(
              (linear): Linear(in_features=4, out_features=3, bias=True)
              (criterion): CrossEntropyLoss()
            )
          (optimizer): SGD (
            Parameter Group 0...
                lr: 0.01
                maximize: False
                momentum: 0
                nesterov: False
                weight_decay: 0
            )
          (lr_scheduler): <torch.optim.lr_scheduler.StepLR object at 0x...>
        )
    """
    if isinstance(creator, dict):
        logger.info(
            "Initializing a core engine modules creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseCoreCreator.factory(**creator)
    if not isinstance(creator, BaseCoreCreator):
        logger.warning(f"creator is not a `BaseCoreCreator` (received: {type(creator)})")
    return creator



r"""This module defines the confusion matrix metric for binary
labels."""

from __future__ import annotations

__all__ = ["BinaryConfusionMatrix", "CategoricalConfusionMatrix"]

import logging
from collections.abc import Sequence

from coola.utils import str_mapping
from torch import Tensor

from gravitorch.engines.base import BaseEngine
from gravitorch.models.metrics.base import EmptyMetricError
from gravitorch.models.metrics.base_epoch import BaseEpochMetric
from gravitorch.nn import ToCategoricalLabel
from gravitorch.utils.artifacts import PyTorchArtifact
from gravitorch.utils.exp_trackers import EpochStep
from gravitorch.utils.format import str_scalar
from gravitorch.utils.history import MaxScalarHistory, MinScalarHistory
from gravitorch.utils.meters import BinaryConfusionMatrix as BinaryConfusionMatrixMeter
from gravitorch.utils.meters import (
    MulticlassConfusionMatrix as CategoricalConfusionMatrixMeter,
)
from gravitorch.utils.tensor import str_full_tensor

logger = logging.getLogger(__name__)


class BinaryConfusionMatrix(BaseEpochMetric):
    r"""Implements the confusion matrix metric for binary labels.

    Args:
    ----
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric.
            The name is used to log the metric results.
                Default: ``'bin_conf_mat'``
        betas (sequence, optional): Specifies the betas used to
            compute the f-beta score. Default: ``(1,)``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.models.metrics import BinaryConfusionMatrix
        >>> metric = BinaryConfusionMatrix("eval")
        >>> metric
        BinaryConfusionMatrix(
          (mode): eval
          (name): bin_conf_mat
          (betas): (1,)
          (confusion_matrix): BinaryConfusionMatrix(num_classes=2, num_predictions=0)
        )
        >>> metric(torch.tensor([0, 1, 0, 1]), torch.tensor([0, 1, 0, 1]))
        >>> metric.value()
        {'eval/bin_conf_mat_accuracy': 1.0,
         'eval/bin_conf_mat_balanced_accuracy': 1.0,
         'eval/bin_conf_mat_false_negative_rate': 0.0,
         'eval/bin_conf_mat_false_negative': 0,
         'eval/bin_conf_mat_false_positive_rate': 0.0,
         'eval/bin_conf_mat_false_positive': 0,
         'eval/bin_conf_mat_jaccard_index': 1.0,
         'eval/bin_conf_mat_num_predictions': 4,
         'eval/bin_conf_mat_precision': 1.0,
         'eval/bin_conf_mat_recall': 1.0,
         'eval/bin_conf_mat_true_negative_rate': 1.0,
         'eval/bin_conf_mat_true_negative': 2,
         'eval/bin_conf_mat_true_positive_rate': 1.0,
         'eval/bin_conf_mat_true_positive': 2,
         'eval/bin_conf_mat_f1_score': 1.0}
        >>> metric(torch.tensor([1, 0]), torch.tensor([1, 0]))
        >>> metric.value()
        {'eval/bin_conf_mat_accuracy': 1.0,
         'eval/bin_conf_mat_balanced_accuracy': 1.0,
         'eval/bin_conf_mat_false_negative_rate': 0.0,
         'eval/bin_conf_mat_false_negative': 0,
         'eval/bin_conf_mat_false_positive_rate': 0.0,
         'eval/bin_conf_mat_false_positive': 0,
         'eval/bin_conf_mat_jaccard_index': 1.0,
         'eval/bin_conf_mat_num_predictions': 6,
         'eval/bin_conf_mat_precision': 1.0,
         'eval/bin_conf_mat_recall': 1.0,
         'eval/bin_conf_mat_true_negative_rate': 1.0,
         'eval/bin_conf_mat_true_negative': 3,
         'eval/bin_conf_mat_true_positive_rate': 1.0,
         'eval/bin_conf_mat_true_positive': 3,
         'eval/bin_conf_mat_f1_score': 1.0}
        >>> metric.reset()
        >>> metric(torch.tensor([1, 0]), torch.tensor([1, 0]))
        >>> metric.value()
        {'eval/bin_conf_mat_accuracy': 1.0,
         'eval/bin_conf_mat_balanced_accuracy': 1.0,
         'eval/bin_conf_mat_false_negative_rate': 0.0,
         'eval/bin_conf_mat_false_negative': 0,
         'eval/bin_conf_mat_false_positive_rate': 0.0,
         'eval/bin_conf_mat_false_positive': 0,
         'eval/bin_conf_mat_jaccard_index': 1.0,
         'eval/bin_conf_mat_num_predictions': 2,
         'eval/bin_conf_mat_precision': 1.0,
         'eval/bin_conf_mat_recall': 1.0,
         'eval/bin_conf_mat_true_negative_rate': 1.0,
         'eval/bin_conf_mat_true_negative': 1,
         'eval/bin_conf_mat_true_positive_rate': 1.0,
         'eval/bin_conf_mat_true_positive': 1,
         'eval/bin_conf_mat_f1_score': 1.0}
    """

    def __init__(
        self,
        mode: str,
        name: str = "bin_conf_mat",
        betas: Sequence[int | float] = (1,),
    ) -> None:
        super().__init__(mode=mode, name=name)
        self._confusion_matrix = BinaryConfusionMatrixMeter()
        self._betas = tuple(betas)
        self.reset()

    def extra_repr(self) -> str:
        return str_mapping(
            {
                "mode": self._mode,
                "name": self._name,
                "betas": self._betas,
                "confusion_matrix": self._confusion_matrix,
            }
        )

    def attach(self, engine: BaseEngine) -> None:
        super().attach(engine)
        trackers = [
            MaxScalarHistory(name=f"{self._metric_name}_accuracy"),
            MaxScalarHistory(name=f"{self._metric_name}_balanced_accuracy"),
            MinScalarHistory(name=f"{self._metric_name}_false_negative_rate"),
            MinScalarHistory(name=f"{self._metric_name}_false_positive_rate"),
            MaxScalarHistory(name=f"{self._metric_name}_jaccard_index"),
            MaxScalarHistory(name=f"{self._metric_name}_precision"),
            MaxScalarHistory(name=f"{self._metric_name}_recall"),
            MaxScalarHistory(name=f"{self._metric_name}_true_negative_rate"),
            MaxScalarHistory(name=f"{self._metric_name}_true_positive_rate"),
        ]
        for beta in self._betas:
            trackers.append(MaxScalarHistory(name=f"{self._metric_name}_f{beta}_score"))
        for tracker in trackers:
            engine.add_history(tracker)

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the mean absolute error metric given a mini-batch of
        examples.

        Args:
        ----
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)``
                and type float or long): Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)``
                and type long or float): Specifies the target tensor.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.models.metrics import BinaryConfusionMatrix
            >>> metric = BinaryConfusionMatrix("eval")
            >>> metric(torch.tensor([0, 1, 0, 1]), torch.tensor([0, 1, 0, 1]))
            >>> metric.value()
            {'eval/bin_conf_mat_accuracy': 1.0,
             'eval/bin_conf_mat_balanced_accuracy': 1.0,
             'eval/bin_conf_mat_false_negative_rate': 0.0,
             'eval/bin_conf_mat_false_negative': 0,
             'eval/bin_conf_mat_false_positive_rate': 0.0,
             'eval/bin_conf_mat_false_positive': 0,
             'eval/bin_conf_mat_jaccard_index': 1.0,
             'eval/bin_conf_mat_num_predictions': 4,
             'eval/bin_conf_mat_precision': 1.0,
             'eval/bin_conf_mat_recall': 1.0,
             'eval/bin_conf_mat_true_negative_rate': 1.0,
             'eval/bin_conf_mat_true_negative': 2,
             'eval/bin_conf_mat_true_positive_rate': 1.0,
             'eval/bin_conf_mat_true_positive': 2,
             'eval/bin_conf_mat_f1_score': 1.0}
        """
        self._confusion_matrix.update(prediction.flatten(), target.flatten())

    def reset(self) -> None:
        self._confusion_matrix.reset()

    def value(self, engine: BaseEngine | None = None) -> dict:
        self._confusion_matrix.all_reduce()
        num_predictions = self._confusion_matrix.num_predictions
        if not num_predictions:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        logger.info(f"Confusion matrix\n{str_full_tensor(self._confusion_matrix.matrix)}\n")
        results = self._confusion_matrix.compute_all_metrics(
            self._betas, prefix=f"{self._metric_name}_"
        )
        results[f"{self._metric_name}_num_predictions"] = num_predictions
        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")

        if engine:
            engine.log_metrics(results, EpochStep(engine.epoch))
            engine.create_artifact(
                PyTorchArtifact(
                    tag=f"metric/{self._metric_name}",
                    data={"confusion_matrix": self._confusion_matrix.matrix, "results": results},
                )
            )
        return results


class CategoricalConfusionMatrix(BaseEpochMetric):
    r"""Implements the confusion matrix metric for categorical labels.

    Args:
    ----
        mode (str): Specifies the mode (e.g. train or eval).
        num_classes (int): Specifies the number of classes.
        name (str, optional): Specifies the name of the metric.
            The name is used to log the metric results.
            Default: ``'cat_conf_mat'``
        betas (sequence, optional): Specifies the betas used to
            compute the f-beta score. Default: ``(1,)``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.models.metrics import CategoricalConfusionMatrix
        >>> metric = CategoricalConfusionMatrix("eval", num_classes=3)
        >>> metric
        CategoricalConfusionMatrix(
          (mode): eval
          (name): cat_conf_mat
          (confusion_matrix): MulticlassConfusionMatrix(num_classes=3, num_predictions=0)
          (prediction_transform): ToCategoricalLabel()
        )
        >>> metric(
        ...     torch.tensor([[3, 2, 1], [1, 3, 2], [3, 2, 1], [1, 3, 2]]),
        ...     torch.tensor([0, 1, 0, 1]),
        ... )
        >>> metric.value()  # doctest:+ELLIPSIS
        {'eval/cat_conf_mat_accuracy': 1.0,
         'eval/cat_conf_mat_balanced_accuracy': 0.666666...,
         'eval/cat_conf_mat_macro_precision': 0.666666...,
         'eval/cat_conf_mat_macro_recall': 0.666666...,
         'eval/cat_conf_mat_macro_f1_score': nan,
         'eval/cat_conf_mat_micro_precision': 1.0,
         'eval/cat_conf_mat_micro_recall': 1.0,
         'eval/cat_conf_mat_micro_f1_score': 1.0,
         'eval/cat_conf_mat_weighted_precision': 1.0,
         'eval/cat_conf_mat_weighted_recall': 1.0,
         'eval/cat_conf_mat_weighted_f1_score': nan,
         'eval/cat_conf_mat_num_predictions': 4}
        >>> metric(torch.tensor([[1, 2, 3], [3, 2, 1]]), torch.tensor([2, 0]))
        >>> metric.value()  # doctest:+ELLIPSIS
        {'eval/cat_conf_mat_accuracy': 1.0,
         'eval/cat_conf_mat_balanced_accuracy': 1.0,
         'eval/cat_conf_mat_macro_precision': 1.0,
         'eval/cat_conf_mat_macro_recall': 1.0,
         'eval/cat_conf_mat_macro_f1_score': 1.0,
         'eval/cat_conf_mat_micro_precision': 1.0,
         'eval/cat_conf_mat_micro_recall': 1.0,
         'eval/cat_conf_mat_micro_f1_score': 1.0,
         'eval/cat_conf_mat_weighted_precision': 1.0,
         'eval/cat_conf_mat_weighted_recall': 1.0,
         'eval/cat_conf_mat_weighted_f1_score': 1.0,
         'eval/cat_conf_mat_num_predictions': 6}
        >>> metric.reset()
        >>> metric(torch.tensor([[1, 2, 3], [3, 2, 1]]), torch.tensor([2, 0]))
        >>> metric.value()  # doctest:+ELLIPSIS
        {'eval/cat_conf_mat_accuracy': 1.0,
         'eval/cat_conf_mat_balanced_accuracy': 0.666666...,
         'eval/cat_conf_mat_macro_precision': 0.666666...,
         'eval/cat_conf_mat_macro_recall': 0.666666...,
         'eval/cat_conf_mat_macro_f1_score': nan,
         'eval/cat_conf_mat_micro_precision': 1.0,
         'eval/cat_conf_mat_micro_recall': 1.0,
         'eval/cat_conf_mat_micro_f1_score': 1.0,
         'eval/cat_conf_mat_weighted_precision': 1.0,
         'eval/cat_conf_mat_weighted_recall': 1.0,
         'eval/cat_conf_mat_weighted_f1_score': nan,
         'eval/cat_conf_mat_num_predictions': 2}
    """

    def __init__(
        self,
        mode: str,
        num_classes: int,
        name: str = "cat_conf_mat",
        betas: Sequence[int | float] = (1,),
    ) -> None:
        super().__init__(mode=mode, name=name)
        self.prediction_transform = ToCategoricalLabel()

        self._confusion_matrix = CategoricalConfusionMatrixMeter.from_num_classes(num_classes)
        self._betas = tuple(betas)
        self.reset()

    def extra_repr(self) -> str:
        return str_mapping(
            {
                "mode": self._mode,
                "name": self._name,
                "confusion_matrix": self._confusion_matrix,
            }
        )

    def attach(self, engine: BaseEngine) -> None:
        super().attach(engine)
        trackers = [
            MaxScalarHistory(name=f"{self._metric_name}_accuracy"),
            MaxScalarHistory(name=f"{self._metric_name}_balanced_accuracy"),
            MaxScalarHistory(name=f"{self._metric_name}_macro_precision"),
            MaxScalarHistory(name=f"{self._metric_name}_macro_recall"),
            MaxScalarHistory(name=f"{self._metric_name}_micro_precision"),
            MaxScalarHistory(name=f"{self._metric_name}_micro_recall"),
            MaxScalarHistory(name=f"{self._metric_name}_weighted_precision"),
            MaxScalarHistory(name=f"{self._metric_name}_weighted_recall"),
        ]
        for beta in self._betas:
            trackers.append(MaxScalarHistory(name=f"{self._metric_name}_macro_f{beta}_score"))
            trackers.append(MaxScalarHistory(name=f"{self._metric_name}_micro_f{beta}_score"))
            trackers.append(MaxScalarHistory(name=f"{self._metric_name}_weighted_f{beta}_score"))
        for tracker in trackers:
            engine.add_history(tracker)

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the mean absolute error metric given a mini-batch of
        examples.

        Args:
        ----
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn, num_clasees)`` and type float):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type long or float):
                Specifies the target tensor. The values have to be in
                ``{0, 1, ..., num_classes-1}``.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.models.metrics import CategoricalConfusionMatrix
            >>> metric = CategoricalConfusionMatrix("eval", num_classes=3)
            >>> metric(
            ...     torch.tensor([[3, 2, 1], [1, 3, 2], [3, 2, 1], [1, 3, 2]]),
            ...     torch.tensor([0, 1, 0, 1]),
            ... )
            >>> metric.value()  # doctest:+ELLIPSIS
            {'eval/cat_conf_mat_accuracy': 1.0,
             'eval/cat_conf_mat_balanced_accuracy': 0.666666...,
             'eval/cat_conf_mat_macro_precision': 0.666666...,
             'eval/cat_conf_mat_macro_recall': 0.666666...,
             'eval/cat_conf_mat_macro_f1_score': nan,
             'eval/cat_conf_mat_micro_precision': 1.0,
             'eval/cat_conf_mat_micro_recall': 1.0,
             'eval/cat_conf_mat_micro_f1_score': 1.0,
             'eval/cat_conf_mat_weighted_precision': 1.0,
             'eval/cat_conf_mat_weighted_recall': 1.0,
             'eval/cat_conf_mat_weighted_f1_score': nan,
             'eval/cat_conf_mat_num_predictions': 4}
        """
        self._confusion_matrix.update(
            self.prediction_transform(prediction).flatten(), target.flatten()
        )

    def reset(self) -> None:
        self._confusion_matrix.reset()

    def value(self, engine: BaseEngine | None = None) -> dict:
        self._confusion_matrix.all_reduce()
        num_predictions = self._confusion_matrix.num_predictions
        if not num_predictions:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        logger.info(f"Confusion matrix\n{str_full_tensor(self._confusion_matrix.matrix)}\n")
        results = self._confusion_matrix.compute_scalar_metrics(
            self._betas, prefix=f"{self._metric_name}_"
        )
        results[f"{self._metric_name}_num_predictions"] = num_predictions
        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")

        if engine:
            engine.log_metrics(results, EpochStep(engine.epoch))
            engine.create_artifact(
                PyTorchArtifact(
                    tag=f"metric/{self._metric_name}",
                    data={
                        "confusion_matrix": self._confusion_matrix.matrix,
                        "results": results,
                        "per_class_metrics": self._confusion_matrix.compute_per_class_metrics(),
                    },
                )
            )
        return results



__all__ = [
    "BaseModelCreator",
    "CompiledModelCreator",
    "DataDistributedParallelModelCreator",
    "ModelCreator",
    "is_model_creator_config",
    "setup_model_creator",
]

from gravitorch.creators.model.base import (
    BaseModelCreator,
    is_model_creator_config,
    setup_model_creator,
)
from gravitorch.creators.model.compiled import CompiledModelCreator
from gravitorch.creators.model.ddp import DataDistributedParallelModelCreator
from gravitorch.creators.model.vanilla import ModelCreator



from __future__ import annotations

__all__ = [
    "BaseDataCreator",
    "CacheDataCreator",
    "DataCreator",
    "HypercubeVertexDataCreator",
    "is_datacreator_config",
    "setup_datacreator",
]

from gravitorch.data.datacreators.base import (
    BaseDataCreator,
    is_datacreator_config,
    setup_datacreator,
)
from gravitorch.data.datacreators.caching import CacheDataCreator
from gravitorch.data.datacreators.hypercube import HypercubeVertexDataCreator
from gravitorch.data.datacreators.vanilla import DataCreator



r"""This module defines the squared logarithmic error (SLE) metric."""

from __future__ import annotations

__all__ = ["SquaredAsinhError", "SquaredLogError", "SquaredSymlogError"]

import logging

from torch import Tensor

from gravitorch.models.metrics.base_epoch import BaseStateEpochMetric
from gravitorch.models.metrics.state import BaseState, ErrorState
from gravitorch.nn.functional import asinh_mse_loss, msle_loss, symlog_mse_loss

logger = logging.getLogger(__name__)


class SquaredLogError(BaseStateEpochMetric):
    r"""Implements the squared logarithmic error (SLE) metric.

    This metric is best to use when targets having exponential growth,
    such as population counts, average sales of a commodity over a
    span of years etc. Note that this metric penalizes an
    under-predicted estimate greater than an over-predicted estimate.

    Note: this metric only works with positive value (0 included).

    Args:
    ----
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'sq_log_err'``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.models.metrics import SquaredLogError
        >>> metric = SquaredLogError("eval")
        >>> metric
        SquaredLogError(
          (mode): eval
          (name): sq_log_err
          (state): ErrorState(num_predictions=0)
        )
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()
        {'eval/sq_log_err_mean': 0.0,
         'eval/sq_log_err_min': 0.0,
         'eval/sq_log_err_max': 0.0,
         'eval/sq_log_err_sum': 0.0,
         'eval/sq_log_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()  # doctest: +ELLIPSIS
        {'eval/sq_log_err_mean': 0.080075...,
         'eval/sq_log_err_min': 0.0,
         'eval/sq_log_err_max': 0.480453...,
         'eval/sq_log_err_sum': 0.960906...,
         'eval/sq_log_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()  # doctest: +ELLIPSIS
        {'eval/sq_log_err_mean': 0.240226...,
         'eval/sq_log_err_min': 0.0,
         'eval/sq_log_err_max': 0.480453...,
         'eval/sq_log_err_sum': 0.960906...,
         'eval/sq_log_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "sq_log_err",
        state: BaseState | dict | None = None,
    ) -> None:
        super().__init__(mode=mode, name=name, state=state or ErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the squared logarithmic error metric given a mini-
        batch of examples.

        Args:
        ----
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.models.metrics import SquaredLogError
            >>> metric = SquaredLogError("eval")
            >>> metric(torch.ones(2, 4), torch.ones(2, 4))
            >>> metric.value()
            {'eval/sq_log_err_mean': 0.0,
             'eval/sq_log_err_min': 0.0,
             'eval/sq_log_err_max': 0.0,
             'eval/sq_log_err_sum': 0.0,
             'eval/sq_log_err_num_predictions': 8}
        """
        self._state.update(msle_loss(prediction, target, reduction="none"))


class SquaredSymlogError(BaseStateEpochMetric):
    r"""Implements a metric to compute the squared error on the symlog
    transformed predictions and targets.

    It is a generalization of mean squared logarithmic error (MSLE)
    that works for real values. The ``symlog`` transformation is used
    instead of ``log1p`` because ``symlog`` works on negative values.

    Args:
    ----
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'sq_symlog_err'``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.models.metrics import SquaredSymlogError
        >>> metric = SquaredSymlogError("eval")
        >>> metric
        SquaredSymlogError(
          (mode): eval
          (name): sq_symlog_err
          (state): ErrorState(num_predictions=0)
        )
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()  # doctest: +ELLIPSIS
        {'eval/sq_symlog_err_mean': 0.0,
         'eval/sq_symlog_err_min': 0.0,
         'eval/sq_symlog_err_max': 0.0,
         'eval/sq_symlog_err_sum': 0.0,
         'eval/sq_symlog_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()  # doctest: +ELLIPSIS
        {'eval/sq_symlog_err_mean': 0.080075...,
         'eval/sq_symlog_err_min': 0.0,
         'eval/sq_symlog_err_max': 0.480453...,
         'eval/sq_symlog_err_sum': 0.960906...,
         'eval/sq_symlog_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()  # doctest: +ELLIPSIS
        {'eval/sq_symlog_err_mean': 0.240226...,
         'eval/sq_symlog_err_min': 0.0,
         'eval/sq_symlog_err_max': 0.480453...,
         'eval/sq_symlog_err_sum': 0.960906...,
         'eval/sq_symlog_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "sq_symlog_err",
        state: BaseState | dict | None = None,
    ) -> None:
        super().__init__(mode=mode, name=name, state=state or ErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the squared error on the symlog transformed
        predictions and targets given a mini-batch of examples.

        Args:
        ----
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.models.metrics import SquaredSymlogError
            >>> metric = SquaredSymlogError("eval")
            >>> metric(torch.ones(2, 4), torch.ones(2, 4))
            >>> metric.value()  # doctest: +ELLIPSIS
            {'eval/sq_symlog_err_mean': 0.0,
             'eval/sq_symlog_err_min': 0.0,
             'eval/sq_symlog_err_max': 0.0,
             'eval/sq_symlog_err_sum': 0.0,
             'eval/sq_symlog_err_num_predictions': 8}
        """
        self._state.update(symlog_mse_loss(prediction, target, reduction="none"))


class SquaredAsinhError(BaseStateEpochMetric):
    r"""Implements a metric to compute the squared error on the inverse
    hyperbolic sine (arcsinh) transformed predictions and targets.

    Args:
    ----
        mode (str): Specifies the mode (e.g. train or eval).
        name (str, optional): Specifies the name of the metric. The
            name is used to log the metric results.
            Default: ``'sq_asinh_err'``
        state (``BaseState`` or dict, optional): Specifies the metric
            state or its configuration. If ``None``, ``ErrorState`` is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.models.metrics import SquaredAsinhError
        >>> metric = SquaredAsinhError("eval")
        >>> metric
        SquaredAsinhError(
          (mode): eval
          (name): sq_asinh_err
          (state): ErrorState(num_predictions=0)
        )
        >>> metric(torch.ones(2, 4), torch.ones(2, 4))
        >>> metric.value()  # doctest: +ELLIPSIS
        {'eval/sq_asinh_err_mean': 0.0,
         'eval/sq_asinh_err_min': 0.0,
         'eval/sq_asinh_err_max': 0.0,
         'eval/sq_asinh_err_sum': 0.0,
         'eval/sq_asinh_err_num_predictions': 8}
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()  # doctest: +ELLIPSIS
        {'eval/sq_asinh_err_mean': 0.129469...,
         'eval/sq_asinh_err_min': 0.0,
         'eval/sq_asinh_err_max': 0.776819...,
         'eval/sq_asinh_err_sum': 1.553638...,
         'eval/sq_asinh_err_num_predictions': 12}
        >>> metric.reset()
        >>> metric(torch.eye(2), torch.ones(2, 2))
        >>> metric.value()  # doctest: +ELLIPSIS
        {'eval/sq_asinh_err_mean': 0.388409...,
         'eval/sq_asinh_err_min': 0.0,
         'eval/sq_asinh_err_max': 0.776819...,
         'eval/sq_asinh_err_sum': 1.553638...,
         'eval/sq_asinh_err_num_predictions': 4}
    """

    def __init__(
        self,
        mode: str,
        name: str = "sq_asinh_err",
        state: BaseState | dict | None = None,
    ) -> None:
        super().__init__(mode=mode, name=name, state=state or ErrorState())

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the squared error on the inverse hyperbolic sine
        (arcsinh) transformed predictions and targets given a mini-batch
        of examples.

        Args:
        ----
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` and type float or long):
                Specifies the target tensor.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.models.metrics import SquaredAsinhError
            >>> metric = SquaredAsinhError("eval")
            >>> metric(torch.ones(2, 4), torch.ones(2, 4))
            >>> metric.value()  # doctest: +ELLIPSIS
            {'eval/sq_asinh_err_mean': 0.0,
             'eval/sq_asinh_err_min': 0.0,
             'eval/sq_asinh_err_max': 0.0,
             'eval/sq_asinh_err_sum': 0.0,
             'eval/sq_asinh_err_num_predictions': 8}
        """
        self._state.update(asinh_mse_loss(prediction, target, reduction="none"))



from __future__ import annotations

__all__ = ["NoOptimizerCreator"]

import logging
from typing import TYPE_CHECKING

from torch.nn import Module

from gravitorch.creators.optimizer.base import BaseOptimizerCreator

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class NoOptimizerCreator(BaseOptimizerCreator):
    r"""Implements a no optimizer creator.

    This optimizer creator should be used if you do not want to create
    an optimizer. For example if you only want to evaluate your model,
    you do not need to create an optimizer. The ``create`` method always
    returns ``None``.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine, DummyClassificationModel
        >>> from gravitorch.creators.optimizer import OptimizerCreator
        >>> creator = NoOptimizerCreator()
        >>> creator
        NoOptimizerCreator()
        >>> engine = create_dummy_engine()
        >>> model = DummyClassificationModel()
        >>> optimizer = creator.create(engine, model)
        >>> optimizer
        None
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def create(self, engine: BaseEngine, model: Module) -> None:
        logger.info("No optimizer")
        return



from __future__ import annotations

__all__ = ["BaseRunner"]

from abc import ABC, abstractmethod
from typing import Any

from objectory import AbstractFactory


class BaseRunner(ABC, metaclass=AbstractFactory):
    r"""Defines the base class of the runners.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.runners import TrainingRunner
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> runner = TrainingRunner(engine)
        >>> runner.run()
    """

    @abstractmethod
    def run(self) -> Any:
        r"""Executes the logic of the runner.

        Returns
        -------
            Any artifact of the runner

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.runners import TrainingRunner
            >>> from gravitorch.testing import create_dummy_engine
            >>> runner = TrainingRunner(create_dummy_engine())
            >>> engine = runner.run()
        """



from __future__ import annotations

__all__ = ["OptimizerCreator"]

import logging
from typing import TYPE_CHECKING

from torch.nn import Module
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.optimizers.factory import setup_optimizer

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class OptimizerCreator(BaseOptimizerCreator):
    r"""Implements a vanilla optimizer creator.

    Args:
    ----
        optimizer_config (dict or ``None``, optional): Specifies the
            optimizer configuration. If ``None``, no optimizer is
            created and ``None`` will be returned by the ``create``
            method. Default: ``None``
        add_module_to_engine (bool, optional): If ``True``, the
            optimizer is added to the engine state, so the optimizer
            state is stored when the engine creates a checkpoint.
            Default: ``True``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine, DummyClassificationModel
        >>> from gravitorch.creators.optimizer import OptimizerCreator
        >>> creator = OptimizerCreator({"_target_": "torch.optim.SGD", "lr": 0.01})
        >>> creator
        OptimizerCreator(add_module_to_engine=True)
        >>> engine = create_dummy_engine()
        >>> model = DummyClassificationModel()
        >>> optimizer = creator.create(engine, model)
        >>> optimizer  # doctest: +ELLIPSIS
        SGD (
        Parameter Group 0...
            lr: 0.01
            maximize: False
            momentum: 0
            nesterov: False
            weight_decay: 0
        )
    """

    def __init__(
        self, optimizer_config: dict | None = None, add_module_to_engine: bool = True
    ) -> None:
        self._optimizer_config = optimizer_config
        self._add_module_to_engine = bool(add_module_to_engine)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(add_module_to_engine={self._add_module_to_engine})"

    def create(self, engine: BaseEngine, model: Module) -> Optimizer | None:
        optimizer = setup_optimizer(model=model, optimizer=self._optimizer_config)
        logger.info(f"optimizer:\n{optimizer}")
        if self._add_module_to_engine and optimizer is not None:
            logger.info(f"Adding an optimizer to the engine (key: {ct.OPTIMIZER})...")
            engine.add_module(ct.OPTIMIZER, optimizer)
        return optimizer



from __future__ import annotations

__all__ = ["ChainedDataPipeCreator"]

from collections.abc import Sequence
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_sequence
from torch.utils.data import IterDataPipe

from gravitorch.creators.datapipe.base import BaseDataPipeCreator
from gravitorch.datapipes import create_chained_datapipe

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class ChainedDataPipeCreator(BaseDataPipeCreator):
    r"""Implements a ``DataPipe`` creator to create a sequence of
    ``DataPipe``s from their configuration.

    Args:
    ----
        config (dict or sequence of dict): Specifies the configuration
            of the ``DataPipe`` object to create. See description
            of the ``create_chained_datapipe`` function to
            learn more about the expected values.

    Raises:
    ------
        ValueError if the ``DataPipe`` configuration sequence is
            empty.

    Example usage:

    .. code-block:: pycon

        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.creators.datapipe import ChainedDataPipeCreator
        >>> # Create an IterDataPipe object using a single IterDataPipe object and no source input
        >>> creator = ChainedDataPipeCreator(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...         "iterable": [1, 2, 3, 4],
        ...     }
        ... )
        >>> datapipe = creator.create()
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        >>> # Equivalent to
        >>> creator = ChainedDataPipeCreator(
        ...     [
        ...         {
        ...             "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...             "iterable": [1, 2, 3, 4],
        ...         },
        ...     ]
        ... )
        >>> datapipe = creator.create()
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        >>> # It is possible to use the source_inputs to create the same IterDataPipe object.
        >>> # The data is given by the source_inputs
        >>> creator = ChainedDataPipeCreator(
        ...     config={"_target_": "torch.utils.data.datapipes.iter.IterableWrapper"},
        ... )
        >>> datapipe = creator.create(source_inputs=([1, 2, 3, 4],))
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        >>> # Create an IterDataPipe object using two IterDataPipe objects and no source input
        >>> creator = ChainedDataPipeCreator(
        ...     [
        ...         {
        ...             "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...             "iterable": [1, 2, 3, 4],
        ...         },
        ...         {
        ...             "_target_": "torch.utils.data.datapipes.iter.Batcher",
        ...             "batch_size": 2,
        ...         },
        ...     ]
        ... )
        >>> datapipe = creator.create()
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
        >>> # It is possible to use the source_inputs to create the same IterDataPipe object.
        >>> # A source IterDataPipe object is specified by using source_inputs
        >>> creator = ChainedDataPipeCreator(
        ...     config=[
        ...         {
        ...             "_target_": "torch.utils.data.datapipes.iter.Batcher",
        ...             "batch_size": 2,
        ...         },
        ...     ],
        ... )
        >>> datapipe = creator.create(source_inputs=[IterableWrapper([1, 2, 3, 4])])
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
        >>> # It is possible to create a sequential ``IterDataPipe`` object that takes several
        >>> # IterDataPipe objects as input.
        >>> creator = ChainedDataPipeCreator(
        ...     config=[
        ...         {"_target_": "torch.utils.data.datapipes.iter.Multiplexer"},
        ...         {
        ...             "_target_": "torch.utils.data.datapipes.iter.Batcher",
        ...             "batch_size": 2,
        ...         },
        ...     ],
        ... )
        >>> datapipe = creator.create(
        ...     source_inputs=[
        ...         IterableWrapper([1, 2, 3, 4]),
        ...         IterableWrapper([11, 12, 13, 14]),
        ...     ],
        ... )
        >>> tuple(datapipe)
        ([1, 11], [2, 12], [3, 13], [4, 14])
    """

    def __init__(self, config: dict | Sequence[dict]) -> None:
        if not config:
            raise ValueError("It is not possible to create a DataPipe because config is empty")
        self._config = config

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_sequence(self._config))}\n)"

    def create(
        self, engine: BaseEngine | None = None, source_inputs: Sequence | None = None
    ) -> IterDataPipe:
        return create_chained_datapipe(config=self._config, source_inputs=source_inputs)



from __future__ import annotations

__all__ = ["BasePartitioner", "is_partitioner_config", "setup_partitioner"]

import logging
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)

T = TypeVar("T")


class BasePartitioner(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to implement a partitioner.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.partitioners import FixedSizePartitioner
        >>> partitioner = FixedSizePartitioner(partition_size=3)
        >>> partitioner
        FixedSizePartitioner(partition_size=3, drop_last=False)
        >>> partitions = partitioner.partition(list(range(10)))
        >>> partitions
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
    """

    @abstractmethod
    def partition(self, items: Sequence[T], engine: BaseEngine | None = None) -> list[Sequence[T]]:
        r"""Creates data.

        Args:
        ----
            items: Specifies the sequence to partition.
            engine (``BaseEngine`` or ``None``): Specifies an engine.
                Default: ``None``

        Return:
        ------
            list: The list of partitions.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.data.partitioners import FixedSizePartitioner
            >>> partitioner = FixedSizePartitioner(partition_size=3)
            >>> partitions = partitioner.partition(list(range(10)))
            >>> partitions
            [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
        """


def is_partitioner_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BasePartitioner``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BasePartitioner`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.partitioners import is_partitioner_config
        >>> is_partitioner_config(
        ...     {
        ...         "_target_": "gravitorch.data.partitioners.FixedSizePartitioner",
        ...         "partition_size": 3,
        ...     }
        ... )
        True
    """
    return is_object_config(config, BasePartitioner)


def setup_partitioner(partitioner: BasePartitioner | dict) -> BasePartitioner:
    r"""Sets up a partitioner.

    The partitioner is instantiated from its configuration by using
    the ``BasePartitioner`` factory function.

    Args:
    ----
        partitioner (``BasePartitioner`` or dict): Specifies the
            partitioner or its configuration.

    Returns:
    -------
        ``BasePartitioner``: The instantiated partitioner.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.partitioners import setup_partitioner
        >>> partitioner = setup_partitioner(
        ...     {
        ...         "_target_": "gravitorch.data.partitioners.FixedSizePartitioner",
        ...         "partition_size": 3,
        ...     }
        ... )
        >>> partitioner
        FixedSizePartitioner(partition_size=3, drop_last=False)
    """
    if isinstance(partitioner, dict):
        logger.info(
            "Initializing a partitioner from its configuration... "
            f"{str_target_object(partitioner)}"
        )
        partitioner = BasePartitioner.factory(**partitioner)
    return partitioner



from __future__ import annotations

__all__ = ["BaseLRSchedulerCreator"]

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from torch.optim import Optimizer

from gravitorch.lr_schedulers.base import LRSchedulerType

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class BaseLRSchedulerCreator(ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create a learning rate (LR) scheduler.

    Note that it is not the unique approach to create a LR scheduler.
    Feel free to use other approaches if this approach does not fit your
    needs.

    Example usage:

    .. code-block:: pycon

        >>> import torch
        >>> from gravitorch.testing import create_dummy_engine, DummyClassificationModel
        >>> from gravitorch.creators.lr_scheduler import LRSchedulerCreator
        >>> creator = LRSchedulerCreator(
        ...     {"_target_": "torch.optim.lr_scheduler.StepLR", "step_size": 5}
        ... )
        >>> creator
        LRSchedulerCreator(
          (lr_scheduler_config): {'_target_': 'torch.optim.lr_scheduler.StepLR', 'step_size': 5}
          (lr_scheduler_handler): None
          (add_module_to_engine): True
        )
        >>> engine = create_dummy_engine()
        >>> model = DummyClassificationModel()
        >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
        >>> lr_scheduler = creator.create(engine, optimizer)
        >>> lr_scheduler  # doctest: +ELLIPSIS
        <torch.optim.lr_scheduler.StepLR object at 0x...>
    """

    @abstractmethod
    def create(self, engine: BaseEngine, optimizer: Optimizer | None) -> LRSchedulerType | None:
        r"""Creates an optimizer.

        This method is responsible to register the event handlers
        associated to the LR scheduler. In particular, it should
        register the event to call the ``step`` method of the LR
        scheduler. If the optimizer is ``None``, this function should
        return ``None``  because it does not make sense to define a LR
        scheduler without an optimizer.

        Args:
        ----
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.
            optimizer (``torch.nn.Optimizer``): Specifies the
                optimizer.

        Returns:
        -------
            ``LRSchedulerType`` or ``None``: The created LR scheduler
                or ``None`` if there is no LR scheduler to create.

        Example usage:

        .. code-block:: pycon

            >>> import torch
            >>> from gravitorch.testing import create_dummy_engine, DummyClassificationModel
            >>> from gravitorch.creators.lr_scheduler import LRSchedulerCreator
            >>> creator = LRSchedulerCreator(
            ...     {"_target_": "torch.optim.lr_scheduler.StepLR", "step_size": 5}
            ... )
            >>> engine = create_dummy_engine()
            >>> model = DummyClassificationModel()
            >>> optimizer = torch.optim.SGD(model.parameters(), lr=0.01)
            >>> lr_scheduler = creator.create(engine, optimizer)
            >>> lr_scheduler  # doctest: +ELLIPSIS
            <torch.optim.lr_scheduler.StepLR object at 0x...>
        """



from __future__ import annotations

__all__ = ["EngineEvents", "EpochPeriodicCondition", "IterationPeriodicCondition"]

from typing import TYPE_CHECKING, Any

from minevent import BaseCondition

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class EpochPeriodicCondition(BaseCondition):
    r"""Implements an epoch periodic condition.

    This condition is true every ``freq`` epochs.

    Args:
    ----
        engine (``BaseEngine``): Specifies the engine.
        freq (int): Specifies the frequency.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.engines.events import EpochPeriodicCondition
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> condition = EpochPeriodicCondition(engine, freq=3)
        >>> engine.epoch
        -1
        >>> condition.evaluate()
        False
        >>> engine.increment_epoch()
        >>> condition.evaluate()
        True
        >>> engine.increment_epoch()
        >>> condition.evaluate()
        False
        >>> engine.increment_epoch()
        >>> condition.evaluate()
        False
        >>> engine.increment_epoch()
        >>> condition.evaluate()
        True
    """

    def __init__(self, engine: BaseEngine, freq: int) -> None:
        self._engine = engine
        self._freq = int(freq)

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(freq={self._freq:,}, epoch={self._engine.epoch:,})"

    @property
    def freq(self) -> int:
        r"""``int``: The frequency of the condition."""
        return self._freq

    def evaluate(self) -> bool:
        r"""Evaluates the condition given the current state.

        Returns
        -------
            bool: ``True`` if the condition is ``True``, otherwise
                ``False``.
        """
        return self._engine.epoch % self._freq == 0

    def equal(self, other: Any) -> bool:
        if isinstance(other, EpochPeriodicCondition):
            return self.freq == other.freq
        return False


class IterationPeriodicCondition(BaseCondition):
    r"""Implements an iteration periodic condition.

    This condition is true every ``freq`` iterations.

    Args:
    ----
        engine (``BaseEngine``): Specifies the engine.
        freq (int): Specifies the frequency.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.engines.events import IterationPeriodicCondition
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> condition = IterationPeriodicCondition(engine, freq=3)
        >>> engine.iteration
        -1
        >>> condition.evaluate()
        False
        >>> engine.increment_iteration()
        >>> condition.evaluate()
        True
        >>> engine.increment_iteration()
        >>> condition.evaluate()
        False
        >>> engine.increment_iteration()
        >>> condition.evaluate()
        False
        >>> engine.increment_iteration()
        >>> condition.evaluate()
        True
    """

    def __init__(self, engine: BaseEngine, freq: int) -> None:
        self._engine = engine
        self._freq = int(freq)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(freq={self._freq:,}, "
            f"iteration={self._engine.iteration:,})"
        )

    @property
    def freq(self) -> int:
        r"""``int``: The frequency of the condition."""
        return self._freq

    def evaluate(self) -> bool:
        r"""Evaluates the condition given the current state.

        Returns
        -------
            bool: ``True`` if the condition is ``True``, otherwise ``False``.
        """
        return self._engine.iteration % self._freq == 0

    def equal(self, other: Any) -> bool:
        if isinstance(other, IterationPeriodicCondition):
            return self.freq == other.freq
        return False


class EngineEvents:
    r"""Engine specific event names.

    Every engine should fire these events.
    """

    # Generic. These events are used in both ``train`` and ``eval``
    STARTED: str = "started"
    EPOCH_STARTED: str = "epoch_started"
    EPOCH_COMPLETED: str = "epoch_completed"
    COMPLETED: str = "completed"

    # Train
    TRAIN_STARTED: str = "train_started"
    TRAIN_EPOCH_STARTED: str = "train_epoch_started"
    TRAIN_ITERATION_STARTED: str = "train_iteration_started"
    TRAIN_FORWARD_COMPLETED: str = "train_forward_completed"
    TRAIN_BACKWARD_COMPLETED: str = "train_backward_completed"
    TRAIN_ITERATION_COMPLETED: str = "train_iteration_completed"
    TRAIN_EPOCH_COMPLETED: str = "train_epoch_completed"
    TRAIN_COMPLETED: str = "train_completed"

    # Eval
    EVAL_STARTED: str = "eval_started"
    EVAL_EPOCH_STARTED: str = "eval_epoch_started"
    EVAL_ITERATION_STARTED: str = "eval_iteration_started"
    EVAL_ITERATION_COMPLETED: str = "eval_iteration_completed"
    EVAL_EPOCH_COMPLETED: str = "eval_epoch_completed"
    EVAL_COMPLETED: str = "eval_completed"



from __future__ import annotations

__all__ = ["is_optimizer_creator_config", "setup_optimizer_creator"]

import logging

from objectory.utils import is_object_config

from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.creators.optimizer.noo import NoOptimizerCreator
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def is_optimizer_creator_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseOptimizerCreator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseOptimizerCreator`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.optimizer import is_optimizer_creator_config
        >>> is_optimizer_creator_config(
        ...     {
        ...         "_target_": "gravitorch.creators.optimizer.OptimizerCreator",
        ...         "optimizer_config": {"_target_": "torch.optim.SGD", "lr": 0.01},
        ...     }
        ... )
        True
    """
    return is_object_config(config, BaseOptimizerCreator)


def setup_optimizer_creator(creator: BaseOptimizerCreator | dict | None) -> BaseOptimizerCreator:
    r"""Sets up the optimizer creator.

    The optimizer creator is instantiated from its configuration
    by using the ``BaseOptimizerCreator`` factory function.

    Args:
    ----
        creator (``BaseOptimizerCreator`` or dict or ``None``):
            Specifies the optimizer creator or its configuration.
            If ``None``, a ``NoOptimizerCreator`` is created.

    Returns:
    -------
        ``BaseOptimizerCreator``: The instantiated optimizer creator.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine, DummyClassificationModel
        >>> from gravitorch.creators.optimizer import setup_optimizer_creator
        >>> creator = setup_optimizer_creator(
        ...     {
        ...         "_target_": "gravitorch.creators.optimizer.OptimizerCreator",
        ...         "optimizer_config": {"_target_": "torch.optim.SGD", "lr": 0.01},
        ...     }
        ... )
        >>> creator
        OptimizerCreator(add_module_to_engine=True)
    """
    if creator is None:
        creator = NoOptimizerCreator()
    if isinstance(creator, dict):
        logger.info(
            "Initializing the optimizer creator from its configuration... "
            f"{str_target_object(creator)}"
        )
        creator = BaseOptimizerCreator.factory(**creator)
    if not isinstance(creator, BaseOptimizerCreator):
        logger.warning(f"creator is not a `BaseOptimizerCreator` (received: {type(creator)})")
    return creator



r"""This module implements some utility functions for the evaluation
loops."""

from __future__ import annotations

__all__ = ["is_evaluation_loop_config", "setup_evaluation_loop"]

import logging

from objectory.utils import is_object_config

from gravitorch.loops.evaluation.base import BaseEvaluationLoop
from gravitorch.loops.evaluation.vanilla import EvaluationLoop
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def is_evaluation_loop_config(config: dict) -> bool:
    r"""Indicates if the input configuration is a configuration for a
    ``BaseEvaluationLoop``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseEvaluationLoop`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.loops.evaluation import is_evaluation_loop_config
        >>> is_evaluation_loop_config({"_target_": "gravitorch.loops.evaluation.EvaluationLoop"})
        True
    """
    return is_object_config(config, BaseEvaluationLoop)


def setup_evaluation_loop(evaluation_loop: BaseEvaluationLoop | dict | None) -> BaseEvaluationLoop:
    r"""Sets up the evaluation loop.

    The evaluation loop is instantiated from its configuration by
    using the ``BaseEvaluationLoop`` factory function.

    Args:
    ----
        evaluation_loop (``BaseEvaluationLoop`` or dict or None):
            Specifies the evaluation loop or its configuration.
            If ``None``, the ``EvaluationLoop`` is instantiated.

    Returns:
    -------
        ``BaseEvaluationLoop``: The evaluation loop.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.loops.evaluation import setup_evaluation_loop
        >>> loop = setup_evaluation_loop({"_target_": "gravitorch.loops.evaluation.EvaluationLoop"})
        >>> loop
        EvaluationLoop(
          (batch_device_placement): AutoDevicePlacement(device=cpu)
          (grad_enabled): False
          (tag): eval
          (condition): EveryEpochEvalCondition(every=1)
          (observer): NoOpLoopObserver()
          (profiler): NoOpProfiler()
        )
    """
    if evaluation_loop is None:
        evaluation_loop = EvaluationLoop()
    if isinstance(evaluation_loop, dict):
        logger.info(
            f"Initializing an evaluation loop from its configuration... "
            f"{str_target_object(evaluation_loop)}"
        )
        evaluation_loop = BaseEvaluationLoop.factory(**evaluation_loop)
    if not isinstance(evaluation_loop, BaseEvaluationLoop):
        logger.warning(
            f"evaluation_loop is not a `BaseEvaluationLoop` (received: {type(evaluation_loop)})"
        )
    return evaluation_loop



from __future__ import annotations

__all__ = ["LogCudaMemory", "LogSysInfo"]

import logging
from types import TracebackType

from gravitorch.rsrc.base import BaseResource
from gravitorch.utils.cudamem import log_max_cuda_memory_allocated
from gravitorch.utils.sysinfo import log_system_info

logger = logging.getLogger(__name__)


class LogCudaMemory(BaseResource):
    r"""Implements a context manager to log the CUDA memory.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.rsrc import LogCudaMemory
        >>> with LogCudaMemory():
        ...     pass
        ...
    """

    def __enter__(self) -> LogCudaMemory:
        log_max_cuda_memory_allocated()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        log_max_cuda_memory_allocated()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"


class LogSysInfo(BaseResource):
    r"""Implements a context manager to log system information.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.rsrc import LogSysInfo
        >>> with LogSysInfo():
        ...     pass
        ...
    """

    def __enter__(self) -> LogSysInfo:
        log_system_info()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        log_system_info()

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"



from __future__ import annotations

__all__ = ["DirFilterIterDataPipe", "FileFilterIterDataPipe", "PathListerIterDataPipe"]

from collections.abc import Iterator
from pathlib import Path

from coola.utils import str_indent, str_mapping
from torch.utils.data import IterDataPipe


class DirFilterIterDataPipe(IterDataPipe[Path]):
    r"""Implements an ``IterDataPipe`` to keep only the directory.

    Args:
    ----
        datapipe (``IterDataPipe``): Specifies the source
            ``IterDataPipe``.

    Example usage:

    .. code-block:: pycon

        >>> from pathlib import Path
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.datapipes.iter import DirFilter
        >>> dp = DirFilter(IterableWrapper([Path("tmp/A"), Path("tmp/B"), Path("tmp/C")]))
        >>> dp
        DirFilterIterDataPipe(
          (datapipe): IterableWrapperIterDataPipe
        )
    """

    def __init__(self, datapipe: IterDataPipe[Path]) -> None:
        self._datapipe = datapipe

    def __iter__(self) -> Iterator[Path]:
        for path in self._datapipe:
            if path.is_dir():
                yield path

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        args = str_indent(str_mapping({"datapipe": self._datapipe}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"


class FileFilterIterDataPipe(IterDataPipe[Path]):
    r"""Implements an ``IterDataPipe`` to keep only the files.

    Args:
    ----
        datapipe (``IterDataPipe``): Specifies the source
            ``IterDataPipe``.

    Example usage:

    .. code-block:: pycon

        >>> from pathlib import Path
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.datapipes.iter import FileFilter
        >>> dp = FileFilter(IterableWrapper([Path("tmp/A"), Path("tmp/B"), Path("tmp/C")]))
        >>> dp
        FileFilterIterDataPipe(
          (datapipe): IterableWrapperIterDataPipe
        )
    """

    def __init__(self, datapipe: IterDataPipe[Path]) -> None:
        self._datapipe = datapipe

    def __iter__(self) -> Iterator[Path]:
        for path in self._datapipe:
            if path.is_file():
                yield path

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        args = str_indent(str_mapping({"datapipe": self._datapipe}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"


class PathListerIterDataPipe(IterDataPipe[Path]):
    r"""Implements an ``IterDataPipe`` to list the paths.

    Args:
    ----
        datapipe (``IterDataPipe``): Specifies the source
            ``IterDataPipe`` with the root paths.
        pattern (str, optional): Specifies a glob pattern, to return
            only the matching paths. Default: ``'*'``
        deterministic (bool, optional): If ``True``, the paths are
            returned in a deterministic order. Default: ``True``

    Example usage:

    .. code-block:: pycon

        >>> from pathlib import Path
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.datapipes.iter import PathLister
        >>> dp = PathLister(IterableWrapper([Path("tmp/A"), Path("tmp/B"), Path("tmp/C")]))
        >>> dp
        PathListerIterDataPipe(
          (pattern): *
          (deterministic): True
          (datapipe): IterableWrapperIterDataPipe
        )
    """

    def __init__(
        self,
        datapipe: IterDataPipe[Path],
        pattern: str = "*",
        deterministic: bool = True,
    ) -> None:
        self._datapipe = datapipe
        self._pattern = pattern
        self._deterministic = bool(deterministic)

    def __iter__(self) -> Iterator[Path]:
        for path in self._datapipe:
            paths = path.glob(self._pattern)
            if self._deterministic:
                paths = sorted(paths)
            yield from paths

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "pattern": self._pattern,
                    "deterministic": self._deterministic,
                    "datapipe": self._datapipe,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"



from __future__ import annotations

__all__ = ["ModelCreator"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping
from torch import nn

from gravitorch import constants as ct
from gravitorch.creators.model.base import BaseModelCreator
from gravitorch.models.utils import setup_and_attach_model, setup_model
from gravitorch.utils.device_placement import (
    AutoDevicePlacement,
    BaseDevicePlacement,
    setup_device_placement,
)
from gravitorch.utils.format import str_pretty_json

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class ModelCreator(BaseModelCreator):
    r"""Implements a vanilla model creator.

    This model creator is designed for models that run on a single
    device. If ``device_placement=True``, the device is managed by the
    function ``gravitorch.distributed.device()``.

    Args:
    ----
        model_config (dict): Specifies the model configuration.
        attach_model_to_engine (bool, optional): If ``True``, the
            model is attached to the engine. Default: ``True``
        add_module_to_engine (bool, optional): If ``True``, the model
            is added to the engine state, so the model state is stored
            when the engine creates a checkpoint. Default: ``True``
        device_placement (bool, optional): Specifies the device
            placement module. This module moves the model on a target
            device. If ``None``, an ``AutoDevicePlacement`` object is
            instantiated. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine
        >>> from gravitorch.creators.model import ModelCreator
        >>> creator = ModelCreator({"_target_": "gravitorch.testing.DummyClassificationModel"})
        >>> creator
        ModelCreator(
          (model_config): {'_target_': 'gravitorch.testing.DummyClassificationModel'}
          (attach_model_to_engine): True
          (add_module_to_engine): True
          (device_placement): AutoDevicePlacement(device=cpu)
        )
        >>> engine = create_dummy_engine()
        >>> model = creator.create(engine)
        >>> model
        DummyClassificationModel(
          (linear): Linear(in_features=4, out_features=3, bias=True)
          (criterion): CrossEntropyLoss()
        )
    """

    def __init__(
        self,
        model_config: dict,
        attach_model_to_engine: bool = True,
        add_module_to_engine: bool = True,
        device_placement: BaseDevicePlacement | dict | None = None,
    ) -> None:
        self._model_config = model_config
        self._attach_model_to_engine = bool(attach_model_to_engine)
        self._add_module_to_engine = bool(add_module_to_engine)
        self._device_placement = setup_device_placement(device_placement or AutoDevicePlacement())

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "model_config": str_pretty_json(self._model_config),
                    "attach_model_to_engine": self._attach_model_to_engine,
                    "add_module_to_engine": self._add_module_to_engine,
                    "device_placement": self._device_placement,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def create(self, engine: BaseEngine) -> nn.Module:
        logger.info("Creating model...")
        if self._attach_model_to_engine:
            model = setup_and_attach_model(engine=engine, model=self._model_config)
        else:
            model = setup_model(model=self._model_config)
        model = self._device_placement.send(model)
        if self._add_module_to_engine:
            logger.info(f"Adding a model to the engine state (key: {ct.MODEL})...")
            engine.add_module(ct.MODEL, model)
        return model



from __future__ import annotations

__all__ = ["ModelParameterAnalyzer"]

from typing import TYPE_CHECKING

from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import to_events
from gravitorch.nn.utils.parameter import show_parameter_summary
from gravitorch.utils.events import GEventHandler

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class ModelParameterAnalyzer(BaseHandler):
    r"""Implements a handler to analyze the model parameter values.

    Args:
    ----
        events (str or tuple or list): Specifies the event(s) when to
            analyze the model parameter values. For example, it is
            usually a good idea to look at the model parameters at the
            beginning and end of the training.
            Default: ``('started', 'train_completed')``
        tablefmt (str, optional): Specifies the table format. You can
            find the valid formats at
            https://pypi.org/project/tabulate/.
            Default: ``'fancy_outline'``
        floatfmt (str, optional): Specifies the float format.
            Default: ``'.6f'``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.handlers import ModelParameterAnalyzer
        >>> from gravitorch.testing import create_dummy_engine
        >>> engine = create_dummy_engine()
        >>> handler = ModelParameterAnalyzer()
        >>> handler
        ModelParameterAnalyzer(events=('started', 'train_completed'), tablefmt=fancy_outline, floatfmt=.6f)
        >>> handler.attach(engine)
        >>> engine.fire_event("started")
    """

    def __init__(
        self,
        events: str
        | tuple[str, ...]
        | list[str] = (
            EngineEvents.STARTED,
            EngineEvents.TRAIN_COMPLETED,
        ),
        tablefmt: str = "fancy_outline",
        floatfmt: str = ".6f",
    ) -> None:
        self._events = to_events(events)
        self._tablefmt = str(tablefmt)
        self._floatfmt = str(floatfmt)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(events={self._events}, "
            f"tablefmt={self._tablefmt}, floatfmt={self._floatfmt})"
        )

    def attach(self, engine: BaseEngine) -> None:
        for event in self._events:
            engine.add_event_handler(
                event,
                GEventHandler(self.analyze, handler_kwargs={"engine": engine}),
            )

    def analyze(self, engine: BaseEngine) -> None:
        r"""Analyzes the model parameter values.

        Args:
        ----
            engine (``BaseEngine``): Specifies the engine with the
                model to analyze.

        Example usage:

        .. code-block:: pycon

            >>> from gravitorch.handlers import ModelParameterAnalyzer
            >>> from gravitorch.testing import create_dummy_engine
            >>> engine = create_dummy_engine()
            >>> handler = ModelParameterAnalyzer()
            >>> handler.analyze(engine)
        """
        show_parameter_summary(
            module=engine.model, tablefmt=self._tablefmt, floatfmt=self._floatfmt
        )



from __future__ import annotations

__all__ = [
    "BaseOptimizerCreator",
    "NoOptimizerCreator",
    "OptimizerCreator",
    "ZeroRedundancyOptimizerCreator",
    "is_optimizer_creator_config",
    "setup_optimizer_creator",
]

from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.creators.optimizer.factory import (
    is_optimizer_creator_config,
    setup_optimizer_creator,
)
from gravitorch.creators.optimizer.noo import NoOptimizerCreator
from gravitorch.creators.optimizer.vanilla import OptimizerCreator
from gravitorch.creators.optimizer.zero import ZeroRedundancyOptimizerCreator



from __future__ import annotations

__all__ = ["AdvancedCoreCreator"]

from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping
from torch.nn import Module
from torch.optim import Optimizer

from gravitorch.creators.core.base import BaseCoreCreator
from gravitorch.creators.datasource.base import (
    BaseDataSourceCreator,
    setup_datasource_creator,
)
from gravitorch.creators.lr_scheduler.base import BaseLRSchedulerCreator
from gravitorch.creators.lr_scheduler.factory import setup_lr_scheduler_creator
from gravitorch.creators.model.base import BaseModelCreator, setup_model_creator
from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.creators.optimizer.factory import setup_optimizer_creator
from gravitorch.datasources.base import BaseDataSource
from gravitorch.lr_schedulers.base import LRSchedulerType

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class AdvancedCoreCreator(BaseCoreCreator):
    r"""Implements an advanced core engine moules creator.

    Args:
    ----
        datasource_creator (``BaseDataSourceCreator`` or dict):
            Specifies the datasource creator or its configuration.
        model_creator (``BaseModelCreator`` or dict): Specifies the
            model creator or its configuration.
        optimizer_creator (``BaseOptimizerCreator`` or dict or
            ``None`): Specifies the optimizer creator or its
            configuration. Default: ``None``
        lr_scheduler_creator (``BaseLRSchedulerCreator`` or dict or
            ``None`): Specifies the LR scheduler creator or its
            configuration. Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.creators.core import AdvancedCoreCreator
        >>> from gravitorch.creators.datasource import DataSourceCreator
        >>> from gravitorch.creators.model import ModelCreator
        >>> from gravitorch.creators.optimizer import OptimizerCreator
        >>> from gravitorch.creators.lr_scheduler import LRSchedulerCreator
        >>> from gravitorch.testing import create_dummy_engine
        >>> creator = AdvancedCoreCreator(
        ...     datasource_creator=DataSourceCreator(
        ...         {"_target_": "gravitorch.testing.DummyDataSource"}
        ...     ),
        ...     model_creator=ModelCreator(
        ...         {"_target_": "gravitorch.testing.DummyClassificationModel"}
        ...     ),
        ...     optimizer_creator=OptimizerCreator({"_target_": "torch.optim.SGD", "lr": 0.01}),
        ...     lr_scheduler_creator=LRSchedulerCreator(
        ...         {"_target_": "torch.optim.lr_scheduler.StepLR", "step_size": 5}
        ...     ),
        ... )
        >>> creator  # doctest: +ELLIPSIS
        AdvancedCoreCreator(
          (datasource): DataSourceCreator(
              (config): {'_target_': 'gravitorch.testing.DummyDataSource'}
              (attach_to_engine): True
              (add_module_to_engine): True
            )
          (model_creator): ModelCreator(
              (model_config): {'_target_': 'gravitorch.testing.DummyClassificationModel'}
              (attach_model_to_engine): True
              (add_module_to_engine): True
              (device_placement): AutoDevicePlacement(device=cpu)
            )
          (optimizer_creator): OptimizerCreator(add_module_to_engine=True)
          (lr_scheduler_creator): LRSchedulerCreator(
              (lr_scheduler_config): {'_target_': 'torch.optim.lr_scheduler.StepLR', 'step_size': 5}
              (lr_scheduler_handler): None
              (add_module_to_engine): True
            )
        )
        >>> engine = create_dummy_engine()
        >>> datasource, model, optimizer, lr_scheduler = creator.create(engine)
        >>> datasource
        DummyDataSource(
          (datasets):
            (train): DummyDataset(num_examples=4, feature_size=4)
            (eval): DummyDataset(num_examples=4, feature_size=4)
          (dataloader_creators):
            (train): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
            (eval): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
        )
        >>> model
        DummyClassificationModel(
          (linear): Linear(in_features=4, out_features=3, bias=True)
          (criterion): CrossEntropyLoss()
        )
        >>> optimizer  # doctest: +ELLIPSIS
        SGD (
          Parameter Group 0...
              lr: 0.01
              maximize: False
              momentum: 0
              nesterov: False
              weight_decay: 0
        )
        >>> lr_scheduler  # doctest: +ELLIPSIS
        <torch.optim.lr_scheduler.StepLR object at 0x...>
    """

    def __init__(
        self,
        datasource_creator: BaseDataSourceCreator | dict,
        model_creator: BaseModelCreator | dict,
        optimizer_creator: BaseOptimizerCreator | dict | None = None,
        lr_scheduler_creator: BaseLRSchedulerCreator | dict | None = None,
    ) -> None:
        self._datasource_creator = setup_datasource_creator(datasource_creator)
        self._model_creator = setup_model_creator(model_creator)
        self._optimizer_creator = setup_optimizer_creator(optimizer_creator)
        self._lr_scheduler_creator = setup_lr_scheduler_creator(lr_scheduler_creator)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "datasource": self._datasource_creator,
                    "model_creator": self._model_creator,
                    "optimizer_creator": self._optimizer_creator,
                    "lr_scheduler_creator": self._lr_scheduler_creator,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def create(
        self, engine: BaseEngine
    ) -> tuple[BaseDataSource, Module, Optimizer | None, LRSchedulerType | None]:
        datasource = self._datasource_creator.create(engine=engine)
        model = self._model_creator.create(engine=engine)
        optimizer = self._optimizer_creator.create(engine=engine, model=model)
        lr_scheduler = self._lr_scheduler_creator.create(engine=engine, optimizer=optimizer)
        return datasource, model, optimizer, lr_scheduler



from __future__ import annotations

__all__ = ["SourceWrapperIterDataPipe"]

import copy
import logging
from collections.abc import Iterable, Iterator

from coola import summary
from coola.utils import str_indent, str_mapping
from torch.utils.data import IterDataPipe

logger = logging.getLogger(__name__)


class SourceWrapperIterDataPipe(IterDataPipe):
    r"""Creates a simple source DataPipe from an iterable.

    Based on https://github.com/pytorch/pytorch/blob/3c2199b159b6ec57af3f7ea22d61ace9ce5cf5bc/torch/utils/data/datapipes/iter/utils.py#L8-L50  # noqa: B950

    Args:
    ----
        source (``iterable``): Specifies the input iterable.
        deepcopy (bool, optional): If ``True``, the input iterable
            object is deep-copied before to iterate over the data.
            It allows a deterministic behavior when in-place
            operations are performed on the data. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes.iter import SourceWrapper
        >>> dp = SourceWrapper([1, 2, 3, 4, 5])
        >>> dp
        SourceWrapperIterDataPipe(
          (deepcopy): False
          (source): <class 'list'> (length=5)
                  (0): 1
                  (1): 2
                  (2): 3
                  (3): 4
                  (4): 5
        )
        >>> list(dp)
        [1, 2, 3, 4, 5]
    """

    def __init__(self, source: Iterable, deepcopy: bool = False) -> None:
        self._source = source
        self._deepcopy = bool(deepcopy)

    def __iter__(self) -> Iterator:
        source = self._source
        if self._deepcopy:
            try:
                source = copy.deepcopy(source)
            except TypeError:
                logger.warning(
                    "The input iterable can not be deepcopied, please be aware of in-place "
                    "modification would affect source data."
                )
        yield from source

    def __len__(self) -> int:
        return len(self._source)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "deepcopy": self._deepcopy,
                    "source": str_indent(summary(self._source), num_spaces=4),
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"



from __future__ import annotations

__all__ = ["FixedSizePartitioner"]

from collections.abc import Sequence
from typing import TypeVar

from gravitorch.data.partitioners.base import BasePartitioner
from gravitorch.engines import BaseEngine
from gravitorch.utils.partitioning import fixed_size_partitions

T = TypeVar("T")


class FixedSizePartitioner(BasePartitioner[T]):
    r"""Implements a partitioner that creates fixed-size partitions.

    Args:
    ----
        partition_size (int): Specifies the partition size i.e. the
            number of items in each partition.
        drop_last (bool, optional): If ``True``, it drops the last
            items if the number of items is not evenly divisible by
            ``partition_size``.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.partitioners import FixedSizePartitioner
        >>> partitioner = FixedSizePartitioner(partition_size=3)
        >>> partitioner
        FixedSizePartitioner(partition_size=3, drop_last=False)
        >>> partitions = partitioner.partition(list(range(10)))
        >>> partitions
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
    """

    def __init__(self, partition_size: int, drop_last: bool = False) -> None:
        self._partition_size = int(partition_size)
        self._drop_last = bool(drop_last)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}("
            f"partition_size={self._partition_size:,}, drop_last={self._drop_last})"
        )

    @property
    def drop_last(self) -> bool:
        r"""``bool``: Indicates if the last items are dropped or not if
        there are not enough items."""
        return self._drop_last

    @property
    def partition_size(self) -> int:
        r"""``int``: The partition size."""
        return self._partition_size

    def partition(self, items: Sequence[T], engine: BaseEngine | None = None) -> list[Sequence[T]]:
        return fixed_size_partitions(
            items=items, partition_size=self._partition_size, drop_last=self._drop_last
        )



from __future__ import annotations

__all__ = ["SequentialDataPipeCreator"]

from collections.abc import Sequence
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_sequence
from torch.utils.data import IterDataPipe

from gravitorch.creators.datapipe.base import (
    BaseDataPipeCreator,
    setup_datapipe_creator,
)

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class SequentialDataPipeCreator(BaseDataPipeCreator):
    r"""Implements an ``IterDataPipe`` creator to create an
    ``IterDataPipe`` object by using a sequence ``IterDataPipe``
    creators.

    Args:
    ----
        creators: Specifies the sequence of ``IterDataPipe`` creators
            or their configurations. The sequence of creators follows
            the order of the ``IterDataPipe``s. The first creator is
            used to create the first ``IterDataPipe`` (a.k.a. source),
            and the last creator is used to create the last
            ``IterDataPipe`` (a.k.a. sink). This creator assumes all
            the DataPipes have a single source DataPipe as their first
            argument, excepts for the source ``IterDataPipe``.

    Example usage:

    .. code-block:: pycon

        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> from gravitorch.creators.datapipe import (
        ...     SequentialDataPipeCreator,
        ...     ChainedDataPipeCreator,
        ... )
        >>> # Create an IterDataPipe object using a single IterDataPipe creator and no source input
        >>> creator = SequentialDataPipeCreator(
        ...     [
        ...         ChainedDataPipeCreator(
        ...             {
        ...                 "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...                 "iterable": [1, 2, 3, 4],
        ...             },
        ...         ),
        ...     ]
        ... )
        >>> datapipe = creator.create()
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        >>> # It is possible to use the source_inputs to create the same IterDataPipe object.
        >>> # The data is given by the source_inputs
        >>> creator = SequentialDataPipeCreator(
        ...     [
        ...         ChainedDataPipeCreator(
        ...             {"_target_": "torch.utils.data.datapipes.iter.IterableWrapper"},
        ...         ),
        ...     ]
        ... )
        >>> datapipe = creator.create(source_inputs=([1, 2, 3, 4],))
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        >>> # Create an IterDataPipe object using two IterDataPipe creators and no source input
        >>> creator = SequentialDataPipeCreator(
        ...     [
        ...         ChainedDataPipeCreator(
        ...             {
        ...                 "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...                 "iterable": [1, 2, 3, 4],
        ...             },
        ...         ),
        ...         ChainedDataPipeCreator(
        ...             {
        ...                 "_target_": "torch.utils.data.datapipes.iter.Batcher",
        ...                 "batch_size": 2,
        ...             },
        ...         ),
        ...     ]
        ... )
        >>> datapipe = creator.create()
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
        >>> # It is possible to use the source_inputs to create the same IterDataPipe object.
        >>> # A source IterDataPipe object is specified by using source_inputs
        >>> creator = SequentialDataPipeCreator(
        ...     creators=[
        ...         ChainedDataPipeCreator(
        ...             {
        ...                 "_target_": "torch.utils.data.datapipes.iter.Batcher",
        ...                 "batch_size": 2,
        ...             },
        ...         ),
        ...     ]
        ... )
        >>> datapipe = creator.create(source_inputs=[IterableWrapper([1, 2, 3, 4])])
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
        >>> # It is possible to create a sequential ``IterDataPipe`` object that takes several
        >>> # IterDataPipe objects as input.
        >>> creator = SequentialDataPipeCreator(
        ...     [
        ...         ChainedDataPipeCreator(
        ...             {"_target_": "torch.utils.data.datapipes.iter.Multiplexer"},
        ...         ),
        ...         ChainedDataPipeCreator(
        ...             {
        ...                 "_target_": "torch.utils.data.datapipes.iter.Batcher",
        ...                 "batch_size": 2,
        ...             },
        ...         ),
        ...     ]
        ... )
        >>> datapipe = creator.create(
        ...     source_inputs=[
        ...         IterableWrapper([1, 2, 3, 4]),
        ...         IterableWrapper([11, 12, 13, 14]),
        ...     ],
        ... )
        >>> tuple(datapipe)
        ([1, 11], [2, 12], [3, 13], [4, 14])
    """

    def __init__(self, creators: Sequence[BaseDataPipeCreator | dict]) -> None:
        if not creators:
            raise ValueError("It is not possible to create a DataPipe because creators is empty")
        self._creators = [setup_datapipe_creator(creator) for creator in creators]

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  {str_indent(str_sequence(self._creators))}\n)"

    def create(
        self, engine: BaseEngine | None = None, source_inputs: Sequence | None = None
    ) -> IterDataPipe:
        datapipe = self._creators[0].create(engine=engine, source_inputs=source_inputs)
        for creator in self._creators[1:]:
            datapipe = creator.create(engine=engine, source_inputs=(datapipe,))
        return datapipe



from __future__ import annotations

__all__ = ["DataSourceCreator"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping

from gravitorch import constants as ct
from gravitorch.creators.datasource.base import BaseDataSourceCreator
from gravitorch.datasources.base import (
    BaseDataSource,
    setup_and_attach_datasource,
    setup_datasource,
)
from gravitorch.utils.format import str_pretty_json

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine

logger = logging.getLogger(__name__)


class DataSourceCreator(BaseDataSourceCreator):
    r"""Implements a simple datasource creator.

    Args:
    ----
        config (dict): Specifies the datasource configuration.
        attach_to_engine (bool, optional): If ``True``, the data
            source is attached to the engine. Default: ``True``
        add_module_to_engine (bool, optional): If ``True``, the data
            source is added to the engine state, so the datasource
            state is stored when the engine creates a checkpoint.
            Default: ``True``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine
        >>> from gravitorch.creators.datasource import DataSourceCreator
        >>> creator = DataSourceCreator({"_target_": "gravitorch.testing.DummyDataSource"})
        >>> creator
        DataSourceCreator(
          (config): {'_target_': 'gravitorch.testing.DummyDataSource'}
          (attach_to_engine): True
          (add_module_to_engine): True
        )
        >>> engine = create_dummy_engine()
        >>> datasource = creator.create(engine)
        >>> datasource
        DummyDataSource(
          (datasets):
            (train): DummyDataset(num_examples=4, feature_size=4)
            (eval): DummyDataset(num_examples=4, feature_size=4)
          (dataloader_creators):
            (train): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
            (eval): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
        )
    """

    def __init__(
        self,
        config: dict,
        attach_to_engine: bool = True,
        add_module_to_engine: bool = True,
    ) -> None:
        self._config = config
        self._attach_to_engine = bool(attach_to_engine)
        self._add_module_to_engine = bool(add_module_to_engine)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "config": str_pretty_json(self._config),
                    "attach_to_engine": self._attach_to_engine,
                    "add_module_to_engine": self._add_module_to_engine,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def create(self, engine: BaseEngine) -> BaseDataSource:
        logger.info("Creating a datasource...")
        if self._attach_to_engine:
            datasource = setup_and_attach_datasource(datasource=self._config, engine=engine)
        else:
            datasource = setup_datasource(datasource=self._config)
        if self._add_module_to_engine:
            logger.info(f"Adding a datasource to the engine (key: {ct.DATA_SOURCE})...")
            engine.add_module(ct.DATA_SOURCE, datasource)
        return datasource



__all__ = [
    "AdvancedCoreCreator",
    "BaseCoreCreator",
    "CoreCreator",
    "is_core_creator_config",
    "setup_core_creator",
]

from gravitorch.creators.core.advanced import AdvancedCoreCreator
from gravitorch.creators.core.base import (
    BaseCoreCreator,
    is_core_creator_config,
    setup_core_creator,
)
from gravitorch.creators.core.vanilla import CoreCreator



from __future__ import annotations

__all__ = ["CoreCreator"]

from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping
from torch import nn
from torch.optim import Optimizer

from gravitorch import constants as ct
from gravitorch.creators.core.base import BaseCoreCreator
from gravitorch.datasources.base import BaseDataSource, setup_datasource
from gravitorch.lr_schedulers.base import LRSchedulerType, setup_lr_scheduler
from gravitorch.models.utils import setup_model
from gravitorch.optimizers.factory import setup_optimizer

if TYPE_CHECKING:
    from gravitorch.engines import BaseEngine


class CoreCreator(BaseCoreCreator):
    r"""Implements a simple core engine moules creator.

    This creator does not always "create" the core modules because
    they can already exist. The user is responsible to attach the
    core modules to the engine. This creator only adds the given
    modules to the engine state.

    Args:
    ----
        datasource (``BaseDataSource`` or dict): Specifies the data
            source or its configuration.
        model (``BaseModelCreator`` or dict): Specifies the model
            or its configuration.
        optimizer (``BaseOptimizerCreator`` or dict or ``None`):
            Specifies the optimizer or its configuration.
            Default: ``None``
        lr_scheduler (``BaseLRSchedulerCreator`` or dict or ``None`):
            Specifies the LR scheduler or its configuration.
            Default: ``None``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.testing import create_dummy_engine
        >>> from gravitorch.creators.core import CoreCreator
        >>> creator = CoreCreator(
        ...     datasource={"_target_": "gravitorch.testing.DummyDataSource"},
        ...     model={"_target_": "gravitorch.testing.DummyClassificationModel"},
        ...     optimizer={"_target_": "torch.optim.SGD", "lr": 0.01},
        ...     lr_scheduler={"_target_": "torch.optim.lr_scheduler.StepLR", "step_size": 5},
        ... )
        >>> creator  # doctest: +ELLIPSIS
        CoreCreator(
          (datasource): DummyDataSource(
              (datasets):
                (train): DummyDataset(num_examples=4, feature_size=4)
                (eval): DummyDataset(num_examples=4, feature_size=4)
              (dataloader_creators):
                (train): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
                (eval): DataLoaderCreator(
                    (seed): 0
                    (batch_size): 1
                    (shuffle): False
                  )
            )
          (model): DummyClassificationModel(
              (linear): Linear(in_features=4, out_features=3, bias=True)
              (criterion): CrossEntropyLoss()
            )
          (optimizer): SGD (
            Parameter Group 0...
                lr: 0.01
                maximize: False
                momentum: 0
                nesterov: False
                weight_decay: 0
            )
          (lr_scheduler): <torch.optim.lr_scheduler.StepLR object at 0x...>
        )
        >>> engine = create_dummy_engine()
        >>> datasource, model, optimizer, lr_scheduler = creator.create(engine)
        >>> datasource
        DummyDataSource(
          (datasets):
            (train): DummyDataset(num_examples=4, feature_size=4)
            (eval): DummyDataset(num_examples=4, feature_size=4)
          (dataloader_creators):
            (train): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
            (eval): DataLoaderCreator(
                (seed): 0
                (batch_size): 1
                (shuffle): False
              )
        )
        >>> model
        DummyClassificationModel(
          (linear): Linear(in_features=4, out_features=3, bias=True)
          (criterion): CrossEntropyLoss()
        )
        >>> optimizer  # doctest: +ELLIPSIS
        SGD (
          Parameter Group 0...
              lr: 0.01
              maximize: False
              momentum: 0
              nesterov: False
              weight_decay: 0
        )
        >>> lr_scheduler  # doctest: +ELLIPSIS
        <torch.optim.lr_scheduler.StepLR object at 0x...>
    """

    def __init__(
        self,
        datasource: BaseDataSource | dict,
        model: nn.Module | dict,
        optimizer: Optimizer | dict | None = None,
        lr_scheduler: LRSchedulerType | dict | None = None,
    ) -> None:
        self._datasource = setup_datasource(datasource)
        self._model = setup_model(model)
        self._optimizer = setup_optimizer(model=self._model, optimizer=optimizer)
        self._lr_scheduler = setup_lr_scheduler(
            optimizer=self._optimizer, lr_scheduler=lr_scheduler
        )

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "datasource": self._datasource,
                    "model": self._model,
                    "optimizer": self._optimizer,
                    "lr_scheduler": self._lr_scheduler,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def create(
        self, engine: BaseEngine
    ) -> tuple[BaseDataSource, nn.Module, Optimizer | None, LRSchedulerType | None]:
        engine.add_module(ct.DATA_SOURCE, self._datasource)
        engine.add_module(ct.MODEL, self._model)
        if self._optimizer:
            engine.add_module(ct.OPTIMIZER, self._optimizer)
        if self._lr_scheduler:
            engine.add_module(ct.LR_SCHEDULER, self._lr_scheduler)
        return self._datasource, self._model, self._optimizer, self._lr_scheduler
